@echo off
REM Created by Tyge Lovset Nov. 2016 (tycho)
REM flac.exe and lossywav.exe are released under the GPL license (see gpl.txt)

:thread_loop
if not "%~1"==":" goto start
for /f "usebackq delims=" %%F in ("%inputfiles%") do (set "s_1=%%~F"& call :thread %2 %args%)
goto :eof

:thread
set proc=%1
shift
set /a filecount+=1
set lockfile="%workdir%\lock_%filecount%.tmp"
if exist "%lockfile%" goto :eof
copy nul "%lockfile%">nul
goto single_file

REM ---------------------- REGULAR START ---------------------------

:start
setlocal
set ver=1.4.0
set numprocs=%NUMBER_OF_PROCESSORS%

REM Auto detect tools
set tools=
if exist "%~dp0lossyWAV.exe" if exist "%~dp0metaflac.exe" if exist "%~dp0flac.exe" set "tools=%~dp0"

echo lossyflac %ver%, Copyright (C) 2016 tycho.
if "%~1"=="" (
  echo.
  echo Usage: %~n0 inputfile... [lossywav-options] [@ flac-options]
  echo        %~n0 folder [-# threads] [lossywav-options] [@ flac-options]
  echo        %~n0 --help
  goto finish
)
if "%~1"=="--help" (
  echo.
  echo Convert between WAV/FLAC and lossy.flac. Resulting flac files contains 
  echo tags and picture art from input flac files. Checking tools:
  echo.
  "%tools%lossyWAV.exe" --version
  "%tools%metaflac.exe" --version
  "%tools%flac.exe" --version
  echo.
  echo Usage: %~nx0 inputfile... [lossywav-options] [@ flac-options]
  echo        %~nx0 folder [-# threads] [lossywav-options] [@ flac-options]
  echo.
  echo Examples
  echo    %~n0 *.flac -q P @ -8                  create lossy.flacs in same dir
  echo    %~n0 *.wav -o lwdir                    create lossy.flac files in lwdir
  echo    %~n0 *.flac -o lwdir -C -f             create lossy+corr. flac files 
  echo    %~n0 lwdir\*.lossy.flac -M -o flacdir  merge lossy+corr files back to lossless
  echo    %~n0 - -C @ -o song.lossy.flac         create lossy+corr. from std input
  echo    %~n0 song.lossy.flac -M                merge back to lossless song.flac 
  echo    %~n0 Music -o Lossy -C                 convert all flacs in Music recursively
  echo    %~n0 Lossy -o Music2 -M -# 1           reconstruct flacs from Lossy recursively
  goto finish
)
set "batch=%~f0"
set files=
set fileargs=0
set recurse=0
set finished=0
set proc=1
:files_loop
  if "%~1"=="" goto files_done
  if "%~1"=="@" goto files_done
  set "f=%~1"
  if not "%~1"=="-" if "%f:~0,1%"=="-" goto files_done
  set files=%files% "%~1"
  set /a fileargs+=1
  shift& goto files_loop
:files_done

if %fileargs%==0 (
  echo ERROR: a filename or "-" must be specified as first argument.
  goto finish
)

set args=
set merge=0
:args_loop
  if "%~1"=="-M" if "%merge%"=="0" set merge=1
  if "%~1"=="--merge" if "%merge%"=="0" set merge=1
  if "%~1"=="@" if "%merge%"=="0" set merge=@
  if "%~1"=="-#" (set numprocs=%~2& shift& shift& goto args_loop)
  if not "%~1"=="" (set args=%args% "%~1"& shift& goto args_loop)
REM done

set "workdir=%TEMP%\lossyflac_%RANDOM%%RANDOM%"
set "inputfiles=%workdir%\inputfiles.txt"
mkdir "%workdir%"
set filecount=0
REM -- Processing of files, single or parallel.
if %fileargs%==1 if exist %files%\ set recurse=1
if %recurse%==0 for %%F in (%files%) do (set "s_1=%%~F"& call :single_file %args%)
if %recurse%==0 goto end_work
  if not "%merge%"=="1" dir /s /b /a-d %files%\*.flac|findstr /v ".lossy.flac"|findstr /v ".lwcdf.flac"> "%inputfiles%"
  if "%merge%"=="1" dir /s /b /a-d %files%\*.lossy.flac > "%inputfiles%"
  for /f %%C in ('find /v /c "" ^< "%inputfiles%"') do set numfiles=%%C
  set parallel=0
  if not "%numprocs%"=="" if %numprocs% GTR 1 if %numfiles% GTR 1 set parallel=1
  if %parallel%==1 goto parallel
    for /f "usebackq delims=" %%F in ("%inputfiles%") do (set "s_1=%%~F"& call :single_file %args%)
    goto end_work
:parallel
    if %numprocs% GTR 16 set numprocs=16
    if %numprocs% GTR %numfiles% set numprocs=%numfiles%
	echo Converting with %numprocs% parallel threads.
    for /l %%P in (2,1,%numprocs%) do start "lossyflac thread %%P" /min "%COMSPEC%" /c @call "%batch%" : %%P
	call :thread_loop : 1
    echo.
    <nul set /p ="Threads to finish."
:waitfinish	
    set lastwins=%wins%
    for /f %%C in ('tasklist /v ^|find /c "lossyflac thread"') do set wins=%%C
    if not "%wins%"=="%lastwins%" <nul set /p ="..%wins%"
    if %wins%==0 (echo.& goto end_work)
    timeout /nobreak /t 4 >nul
    goto waitfinish
:end_work	
rmdir /s /q "%workdir%"
:finish
endlocal
goto :eof


REM ----------------------------- SUBROUTINES --------------------------------

:single_file
if %parallel%==0 set /a filecount+=1
if "%cancel%"=="1" goto :eof
if %recurse%==0 echo File %filecount%:
if %recurse%==1 echo File %filecount% / %numfiles%:
if %proc% NEQ 1 goto end_choice
  choice /n /cs /c cq /d c /t 1 /m "Hit q to quit."
  if %errorlevel%==2 (set cancel=1& goto :eof)
:end_choice
setlocal
set "wd=%workdir%\%proc%"
if not exist "%wd%\" mkdir "%wd%"
set lw_opts=
set fl_opts=
set "input=%s_1%"
for %%F in ("%s_1%") do set "iname=%%~nF"
for %%F in ("%iname%") do set "ibname=%%~nF"
for %%F in ("%s_1%") do set "idir=%%~dpF"
set "outdir=%idir%"
if %recurse%==1 for %%F in (%files%) do set "outdir=%%~F"
set "oname=%iname%"
set output=
set force=
set operation=NORMAL
set is_flac_in=0
for %%F in ("%s_1%") do if /i "%%~xF"==".flac" set is_flac_in=1
for %%F in ("%s_1%") do if /i "%%~xF"==".fla" set is_flac_in=1

:lw_loop
  shift
  if "%~0"=="" goto single_main
  if "%~0"=="@" goto fl_loop
  if "%~0"=="-f" (set force=-f& goto lw_loop)
  if "%~0"=="--force" (set force=-f& goto lw_loop)
  if "%~0"=="-C" set operation=CORRECTION
  if "%~0"=="--correction" set operation=CORRECTION
  if "%~0"=="-M" set operation=MERGE
  if "%~0"=="--merge" set operation=MERGE
  if "%~0"=="-o" (set "outdir=%~1"& shift& goto lw_loop)
  if "%~0"=="--outdir" (set "outdir=%~1"& shift& goto lw_loop)
  set "lw_opts=%lw_opts% %0"
goto lw_loop
:fl_loop
  shift
  if "%~0"=="" goto single_main
  if "%~0"=="-f" (set force=-f& goto fl_loop)
  if "%~0"=="--force" (set force=-f& goto fl_loop)
  if "%~0"=="-o" (set "outdir=%~dp1"& set "oname=%~n1"& set "output=%~1"& shift& goto fl_loop)  
  if "%~0"=="--output-name" (set "outdir=%~dp1"& set "oname=%~n1"& set "output=%~1"& shift& goto fl_loop)
  set "fl_opts=%fl_opts% %0"
goto fl_loop

:single_main
REM strip final backslash
if "%outdir:~-1%"=="\" set "outdir=%outdir:~0,-1%"

REM -- Insert subdirectory in path in recursive mode.
if %recurse%==0 goto end_appendsubdir
  for /d %%D in (%files%\) do set "xfile=%%~fD"
  for %%F in ("%input%") do set "xinput=%%~dpF"
  call set "subdir=%%xinput:%xfile%=%%"
  if not "%output%"=="" for %%F in ("%output%") do set "output=%%~dpF%subdir%%%~nxF"
  set "outdir=%outdir%\%subdir%"
  if "%outdir:~-1%"=="\" set "outdir=%outdir:~0,-1%"
:end_appendsubdir

if not exist "%outdir%\" mkdir "%outdir%"  
if not exist "%outdir%\" (
  echo ERROR: Output folder "%outdir%\" could not be created. Skipping "%input%".
  goto end_single
)

for %%F in ("%oname%") do if /I "%%~xF"==".lossy" set "oname=%%~nF"
if "%input%"=="-" set "iname=%oname%"
if "%output%"=="" (
  if %operation%==MERGE (set "output=%outdir%\%oname%.flac") else (set "output=%outdir%\%oname%.lossy.flac")
)

REM call :debug
call :%operation%
:end_single
endlocal
goto :eof

REM ----------------------------- OPERATIONS --------------------------------

:NORMAL
if %is_flac_in%==1 (
  "%tools%flac.exe" -d "%input%" -o - --silent | "%tools%lossyWAV.exe" - %lw_opts% --stdout --stdinname "%oname%.wav" | "%tools%flac.exe" - -o "%output%" -b 512 --silent %fl_opts% %force%
  call :tags
) else (
  "%tools%lossyWAV.exe" "%input%" %lw_opts% --stdout --stdinname "%oname%.wav" | "%tools%flac.exe" - -o "%output%" -b 512 --silent %fl_opts% %force%
)
goto :eof


:CORRECTION
if "%force%"=="" if exist "%output%" (
  echo ERROR: output file "%output%" already exists. Use -f to override.
  goto :eof
)
if %is_flac_in%==1 (
  "%tools%flac.exe" "%input%" -d -o - --silent | "%tools%lossyWAV.exe" - %lw_opts% --stdinname "%oname%.wav" --outdir "%wd%" --force
) else (
  "%tools%lossyWAV.exe" "%input%" %lw_opts% --stdinname "%oname%.wav" --outdir "%wd%" --force
)
"%tools%flac.exe" "%wd%\%oname%.lossy.wav" --totally-silent --delete-input-file --keep-foreign-metadata -b 512 %fl_opts% --force -o "%output%"
"%tools%flac.exe" "%wd%\%oname%.lwcdf.wav" --totally-silent --delete-input-file --keep-foreign-metadata %fl_opts% --force -o "%outdir%\%oname%.lwcdf.flac"
set "t_3=%outdir%\%oname%.lwcdf.flac"
if %is_flac_in%==1 call :tags
goto :eof


:MERGE
if %is_flac_in%==1 (
  if not exist "%idir%%ibname%.lwcdf.flac" goto :eof
  if "%force%"=="" if exist "%output%" (
    echo ERROR: output file "%output%" already exists. Use -f to override.
    goto :eof
  )
  "%tools%flac.exe" "%input%" -d --totally-silent --keep-foreign-metadata -o "%wd%\%oname%.lossy.wav" --force
  "%tools%flac.exe" "%idir%%ibname%.lwcdf.flac" -d --totally-silent --keep-foreign-metadata -o "%wd%\%oname%.lwcdf.wav" --force

  "%tools%lossyWAV.exe" "%wd%\%oname%.lossy.wav" %lw_opts% --outdir "%wd%" --force
  del "%wd%\%oname%.lossy.wav" "%wd%\%oname%.lwcdf.wav"
  "%tools%flac.exe" "%wd%\%oname%.wav" --delete-input-file %fl_opts% --force -o "%output%"
  call :tags
) else (
  echo ERROR: lossyflac merges flac'ed lossyWAV files only. Use lossyWAV directly for wav files.
)
goto :eof

REM ----------------------------- EXTRA --------------------------------

:tags
set "tag=%wd%\%oname%_tag.txt"
set "pic=%wd%\%oname%_pic.jpg"
set "t_1=%input%"
set "t_2=%output%"
"%tools%metaflac.exe" --export-tags-to="%tag%" --export-picture-to="%pic%" "%t_1%" 2>nul
if exist "%pic%" (
  "%tools%metaflac.exe" --import-tags-from="%tag%" --import-picture-from="%pic%" "%t_2%"
  REM Tag the correction file too. Overkill?
  if not "%t_3%"=="" "%tools%metaflac.exe" --import-tags-from="%tag%" "%t_3%"
  del "%tag%" "%pic%"
) else (
  "%tools%metaflac.exe" --import-tags-from="%tag%" "%t_2%" "%t_3%"
  del "%tag%"
)
set t_3=
goto :eof


:debug
echo input: "%input%"
echo idir: "%idir%"
echo oname: "%oname%"
echo outdir: "%outdir%"
echo output: "%output%"
echo operation: %operation%
echo lw_opts: %lw_opts%
echo fl_opts: %fl_opts%
echo.
goto :eof
